from functools import cached_property
from typing import Any, Optional

import dotenv
from ape_ethereum.transactions import StaticFeeTransaction
from pydantic import Field, PrivateAttr
import os
from google.cloud import kms
from cryptography.hazmat.primitives import serialization
from ape.api.accounts import AccountAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, SignableMessage, TransactionSignature
from eth_account._utils.legacy_transactions import serializable_unsigned_transaction_from_dict
from eth_account.messages import _hash_eip191_message, encode_defunct
from eth_pydantic_types import HexBytes
from eth_typing import Hash32
from eth_utils import keccak, to_checksum_address
import ecdsa
from rich.traceback import install
from web3 import Web3
from rich.console import Console

console = Console()

install()
# Constants
SECP256_K1_N = int("fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141", 16)


def _convert_der_to_rsv(signature: bytes, v_adjustment_factor: int = 0) -> dict:
    """Convert DER signature to RSV format."""
    r, s = ecdsa.util.sigdecode_der(signature, ecdsa.SECP256k1.order)
    v = v_adjustment_factor
    if s > SECP256_K1_N / 2:
        s = SECP256_K1_N - s
    r = r.to_bytes(32, byteorder="big")
    s = s.to_bytes(32, byteorder="big")
    return dict(v=v, r=r, s=s)


class GCPKmsAccount(AccountAPI):
    """Account implementation using Google Cloud KMS."""

    # Public fields
    key_path: str = Field(default="")

    # Private attributes using Pydantic's PrivateAttr
    _client: kms.KeyManagementServiceClient = PrivateAttr()
    _cached_public_key: Optional[bytes] = PrivateAttr(default=None)
    project_id: str = ""
    location_id: str = ""
    key_ring_id: str = ""
    key_id: str = ""

    def __init__(self, **data):
        super().__init__(**data)
        self._client = kms.KeyManagementServiceClient()

        # Load config from environment
        self.project_id = os.getenv("GOOGLE_CLOUD_PROJECT")
        self.location_id = os.getenv("GOOGLE_CLOUD_REGION")
        self.key_ring_id = os.getenv("KEY_RING")
        self.key_id = os.getenv("KEY_NAME")

        missing_vars = []
        for var in ['project_id', 'location_id', 'key_ring_id', 'key_id']:
            if not getattr(self, var):
                missing_vars.append(var)
        if missing_vars:
            raise ValueError(f"Missing required environment variables: {', '.join(missing_vars)}")

        self.key_path = self.get_key_version_path()

    def extract_public_key_bytes(self, pem_str: str) -> bytes:
        """Extract raw public key bytes from PEM format"""
        if not pem_str.startswith('-----BEGIN PUBLIC KEY-----'):
            pem_str = f"-----BEGIN PUBLIC KEY-----\n{pem_str}\n-----END PUBLIC KEY-----"

        pem_bytes = pem_str.encode('utf-8')
        public_key = serialization.load_pem_public_key(pem_bytes)

        # Get public key bytes and return only X and Y coordinates (64 bytes)
        raw_bytes = public_key.public_bytes(
            encoding=serialization.Encoding.X962,
            format=serialization.PublicFormat.UncompressedPoint
        )

        return raw_bytes[-64:]  # Return only the 64 bytes of X,Y coordinates

    def get_key_version_path(self) -> str:
        """Get the full path to the key version in Cloud KMS."""
        return self._client.crypto_key_version_path(
            self.project_id,
            self.location_id,
            self.key_ring_id,
            self.key_id,
            '1'  # Using version 1
        )

    @property
    def public_key(self) -> bytes:
        """Get public key bytes from KMS."""
        if self._cached_public_key is None:
            response = self._client.get_public_key({
                'name': self.key_path
            })
            if not response.pem:
                raise ValueError("No PEM data in response")

            self._cached_public_key = self.extract_public_key_bytes(response.pem)
        return self._cached_public_key

    @cached_property
    def address(self) -> AddressType:
        return to_checksum_address(keccak(self.public_key)[-20:].hex().lower())

    def _sign_raw_hash(self, msghash: HexBytes | Hash32) -> Optional[bytes]:
        """Sign a message hash using KMS."""
        try:
            response = self._client.asymmetric_sign(
                request={
                    'name': self.key_path,
                    'digest': {
                        'sha256': msghash
                    }
                }
            )
            return response.signature
        except Exception as e:
            console.print(f"Signing error: {e}")
            return None

    def sign_raw_msghash(self, msghash: HexBytes | Hash32) -> Optional[MessageSignature]:
        """Sign a raw message hash and return a MessageSignature."""
        if len(msghash) != 32:
            return None
        if not (signature := self._sign_raw_hash(msghash)):
            return None

        msg_sig = MessageSignature(**_convert_der_to_rsv(signature, 27))

        # Try both v values
        if not self.check_signature(msghash, msg_sig):
            msg_sig = MessageSignature(v=msg_sig.v + 1, r=msg_sig.r, s=msg_sig.s)

        return msg_sig

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        """Sign a message using the HSM key."""
        if isinstance(msg, SignableMessage):
            message = msg
        elif isinstance(msg, str):
            if msg.startswith("0x"):
                message = encode_defunct(hexstr=msg)
            else:
                message = encode_defunct(text=msg)
        elif isinstance(msg, bytes):
            message = encode_defunct(primitive=msg)
        else:
            raise TypeError(f"Unsupported message type: {type(msg)}")

        return self.sign_raw_msghash(_hash_eip191_message(message))

    def sign_transaction(self, txn: TransactionAPI, **signer_options) -> Optional[TransactionAPI]:
        """Sign an EIP-155 transaction."""
        unsigned_txn = serializable_unsigned_transaction_from_dict(txn.model_dump()).hash()
        if not (msg_sig := self._sign_raw_hash(unsigned_txn)):
            return None
        # Calculate v value based on chain ID
        v_base = (2 * txn.chain_id + 35) if txn.chain_id else 27
        txn.signature = TransactionSignature(
            **_convert_der_to_rsv(msg_sig, v_base)
        )

        # Try alternative v value if signature check fails
        if not self.check_signature(txn):
            txn.signature = TransactionSignature(
                v=txn.signature.v + 1,
                r=txn.signature.r,
                s=txn.signature.s,
            )

        return txn

def main():
    dotenv.load_dotenv()
    w3 = Web3(Web3.HTTPProvider('http://localhost:8545'))
    account = GCPKmsAccount()

    console.print(f"GCP KMS Account address: {account.address}")

    # 1. Test Message Signing and Verification
    message = "Hello Ethereum!"
    message_hash = encode_defunct(text=message)

    # Sign the message
    signed_message = account.sign_message(message)
    console.print(f"\nSigned message details:")
    console.print(f"R: {signed_message.r.hex()}")
    console.print(f"S: {signed_message.s.hex()}")
    console.print(f"V: {signed_message.v}")
    console.print(f"Full signature: {signed_message.encode_vrs().hex()}")

    # Verify the signature using web3.py
    recovered_address = w3.eth.account.recover_message(
        message_hash,
        vrs=(signed_message.v, signed_message.r, signed_message.s)
    )
    console.print(f"\nSignature verification:")
    console.print(f"Original address: {account.address}")
    console.print(f"Recovered address: {recovered_address}")
    console.print(f"Signature valid: {recovered_address.lower() == account.address.lower()}")

    # 2. Test Transaction Signing and Sending
    console.print("\nTesting transaction signing...")

    # First, fund the account from a test account (using Anvil's default funded account)
    funded_account = w3.eth.account.from_key("0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80")

    # Send some ETH to our GCP KMS account
    fund_tx = {
        'from': funded_account.address,
        'to': account.address,
        'value': w3.to_wei(0.1, 'ether'),
        'gas': 21000,
        'gasPrice': w3.eth.gas_price,
        'nonce': w3.eth.get_transaction_count(funded_account.address),
        'chainId': w3.eth.chain_id
    }

    # Send funding transaction
    signed_fund_tx = w3.eth.account.sign_transaction(fund_tx, funded_account.key)
    fund_tx_hash = w3.eth.send_raw_transaction(signed_fund_tx.raw_transaction)
    fund_receipt = w3.eth.wait_for_transaction_receipt(fund_tx_hash)
    console.print(f"Funded account with 0.1 ETH. TX hash: {fund_receipt['transactionHash'].hex()}")

    # Now create and sign a transaction from our GCP KMS account
    tx = {
        "chain_id": w3.eth.chain_id,
        "nonce": w3.eth.get_transaction_count(account.address),
        "value": w3.to_wei(0.000001, 'ether'),
        "data": "0x00",
        "receiver": "0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
        "type": 0,
        "gas_limit": 1000000,
        "gas_price": 300000000000,
    }

    # Sign the transaction
    signed_tx = account.sign_transaction(StaticFeeTransaction(**tx))
    console.print(f"{signed_tx=}")

    if signed_tx:
        # Send the transaction
        tx_hash = w3.eth.send_raw_transaction(signed_tx.serialize_transaction())
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        console.print(f"\nTransaction successful!")
        console.print(f"Transaction hash: {receipt['transactionHash'].hex()}")
        console.print(f"From: {receipt['from']}")
        console.print(f"To: {receipt['to']}")
        console.print(f"Gas used: {receipt['gasUsed']}")

        # Verify the transaction signature
        tx_data = w3.eth.get_transaction(tx_hash)
        # Get the raw transaction data
        raw_tx = signed_tx.serialize_transaction()
        recovered_address = w3.eth.account.recover_transaction(raw_tx)
        console.print(f"\nTransaction signature verification:")
        console.print(f"Original address: {account.address}")
        console.print(f"Recovered address: {recovered_address}")
        console.print(f"Signature valid: {recovered_address.lower() == account.address.lower()}")
    else:
        console.print("Failed to sign transaction")


if __name__ == "__main__":
    main()



class SignatureError(Exception):
    """
    Raised when there are issues with signing.
    """


class ConfigurationError(Exception):
    """
    Raised when there are issues with the configuration.
    """



import os

from dotenv import load_dotenv
from eth_utils import keccak, to_checksum_address
from google.cloud import kms_v1
from google.cloud.kms_v1 import Digest
from web3 import Web3
from eth_account.messages import encode_defunct
from eth_account import Account
from google.cloud import kms
import base64
import hashlib
from eth_account.messages import encode_defunct
from eth_utils import keccak, to_bytes, to_hex
from coincurve.utils import hex_to_bytes

# Load environment variables from .env file
load_dotenv()

PROJECT_ID = os.environ["GOOGLE_CLOUD_PROJECT"],
LOCATION_ID = os.environ["GOOGLE_CLOUD_REGION"],
KEY_RING_ID = os.environ["KEY_RING"],
KEY_ID = os.environ["KEY_NAME"]


class EthereumHSMSigner:
    def __init__(
        self,
        project_id: str,
        location_id: str,
        key_ring_id: str,
        key_id: str,
        version_id: str
    ):
        self.client = kms.KeyManagementServiceClient()
        self.key_version_name = self.client.crypto_key_version_path(
            project_id, location_id, key_ring_id, key_id, version_id
        )

    def crc32c(self, data: bytes) -> int:
        """Calculate CRC32C checksum"""
        import crcmod
        crc32c_fun = crcmod.predefined.mkPredefinedCrcFun("crc-32c")
        return crc32c_fun(data)

    def prepare_message(self, message: str) -> bytes:
        """Prepare Ethereum message for signing"""
        # Encode message according to EIP-191
        encoded_message = encode_defunct(text=message)
        # Hash the encoded message
        message_hash = keccak(encoded_message.body)
        return message_hash

    def sign_ethereum_message(self, message: str) -> dict:
        """Sign an Ethereum message using GCP HSM"""
        # Prepare the message hash
        message_hash = self.prepare_message(message)

        # Create the sign request
        digest = {"sha256": message_hash}
        digest_crc32c = self.crc32c(message_hash)

        # Sign using GCP KMS
        sign_response = self.client.asymmetric_sign(
            request={
                "name": self.key_version_name,
                "digest": digest,
                "digest_crc32c": digest_crc32c,
            }
        )

        # Verify response integrity
        if not sign_response.verified_digest_crc32c:
            raise Exception("Request corrupted in-transit")
        if not sign_response.name == self.key_version_name:
            raise Exception("Key name mismatch")
        if not sign_response.signature_crc32c == self.crc32c(sign_response.signature):
            raise Exception("Response corrupted in-transit")

        # Convert DER signature to R,S format
        der_sig = sign_response.signature
        r_length = der_sig[3]
        r = int.from_bytes(der_sig[4:4 + r_length], byteorder='big')
        s_length = der_sig[5 + r_length]
        s = int.from_bytes(der_sig[6 + r_length:6 + r_length + s_length], byteorder='big')

        # Format signature components
        r_hex = hex(r)[2:].zfill(64)
        s_hex = hex(s)[2:].zfill(64)

        return {
            'message_hash': to_hex(message_hash),
            'r': r_hex,
            's': s_hex,
            'signature': f"0x{r_hex}{s_hex}"
        }

    def sign_ethereum_transaction(self, transaction_dict: dict) -> dict:
        """Sign an Ethereum transaction using GCP HSM"""
        # Encode transaction according to EIP-155
        # This is a simplified version - you'll need to implement proper RLP encoding
        transaction_bytes = to_bytes(hexstr=transaction_dict['raw'])
        transaction_hash = keccak(transaction_bytes)

        # Rest of the signing process is similar to message signing
        digest = {"sha256": transaction_hash}
        digest_crc32c = self.crc32c(transaction_hash)

        sign_response = self.client.asymmetric_sign(
            request={
                "name": self.key_version_name,
                "digest": digest,
                "digest_crc32c": digest_crc32c,
            }
        )

        # Similar verification and conversion process as message signing
        # You'll need to implement proper signature encoding for transactions

        return {
            'transaction_hash': to_hex(transaction_hash),
            'signature': to_hex(sign_response.signature)
        }


# Example usage:
if __name__ == "__main__":
    signer = EthereumHSMSigner(
        project_id=PROJECT_ID[0],
        location_id=LOCATION_ID[0],
        key_ring_id=KEY_RING_ID[0],
        key_id=KEY_ID,
        version_id="1"
    )

    # Sign a message
    message = "Hello, Ethereum!"
    signature = signer.sign_ethereum_message(message)
    print(f"Message signature: {signature}")

    # Sign a transaction
    transaction = {
        'raw': '0x...',  # Your raw transaction here
        'nonce': 0,
        'gasPrice': 20000000000,
        'gas': 21000,
        'to': '0x...',
        'value': 1000000000000000000,
        'data': ''
    }
    tx_signature = signer.sign_ethereum_transaction(transaction)
    print(f"Transaction signature: {tx_signature}")



# # Example transaction
# tx = {
#     'to': '0xReceiverAddress',
#     'value': web3.to_wei(0.1, 'ether'),
#     'gas': 2000000,
#     'gasPrice': web3.to_wei('50', 'gwei'),
#     'nonce': web3.eth.get_transaction_count(address),
#     'chainId': chain_id
# }
#
# # Sign the transaction
# signed_tx = web3.eth.account.sign_transaction(tx, private_key=signature)
#
# # Send the transaction
# tx_hash = web3.eth.send_raw_transaction(signed_tx.rawTransaction)
# print(f"Transaction sent: {tx_hash.hex()}")



"""Tests for the CLI interface."""
from typing import Generator
from unittest.mock import patch, MagicMock, PropertyMock

import pytest
from typer.testing import CliRunner
from web3_google_hsm._cli import app

@pytest.fixture
def runner() -> CliRunner:
    """
    Creates a CLI test runner for executing commands in isolation.
    This ensures each test has a fresh CLI environment.
    """
    return CliRunner()

@pytest.fixture
def mock_kms_client() -> Generator[MagicMock, None, None]:
    """
    Mocks the Google KMS client to prevent real API calls.
    This includes mocking both the client initialization and key-related methods.
    """
    with patch("google.cloud.kms.KeyManagementServiceClient") as mock:
        # Create a mock instance for method chaining
        client_instance = mock.return_value

        # Mock the key path method to return a consistent path
        client_instance.crypto_key_version_path.return_value = (
            "projects/test-project/locations/test-region/keyRings/test-keyring/"
            "cryptoKeys/test-key/cryptoKeyVersions/1"
        )

        # Mock the public key retrieval
        public_key_response = MagicMock()
        public_key_response.pem = "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAE0hPxTjwIf407JpkjCdf9kwVPvGdMOZUq\nGaVPbV4qdocIUoJlxmWoOQeL/mR28cLrRqgn+Uj8HAoman2lndsp3w==\n-----END PUBLIC KEY-----\n"

        client_instance.get_public_key.return_value = public_key_response

        yield mock

@pytest.fixture
def mock_account_setup(mock_kms_client: MagicMock) -> Generator[MagicMock, None, None]:
    """
    Creates a complete mock for GCPKmsAccount with all necessary properties and methods.
    This prevents real API calls while providing realistic test responses.
    """
    with patch("web3_google_hsm.GCPKmsAccount") as mock_account:
        # Set up the instance that will be returned
        instance = mock_account.return_value

        # Mock the address property with our specific test address
        type(instance).address = PropertyMock(
            return_value="0x0545640A0EcD6FB6ae94766811F30dCDA4746DFC"
        )

        # Mock the key creation classmethod
        mock_key = MagicMock()
        mock_key.name = (
            "projects/test-project/locations/test-region/keyRings/test-keyring/"
            "cryptoKeys/test-key"
        )
        mock_account.create_eth_key.return_value = mock_key

        # Create a realistic signature mock with our specific test values
        mock_signature = MagicMock()
        mock_signature.v = 27
        mock_signature.r = bytes.fromhex(
            "0ef3b04562ace166a974a96ccbaa05c4bfdc85add2a621added6754e75717d9e"
        )
        mock_signature.s = bytes.fromhex(
            "0669929c316bd7a6cd4688526f4fa446ffa25457d9d50cf089e71d111eaed9b0"
        )
        # Combine r and s into the complete signature
        mock_signature.to_hex.return_value = (
            "0x0ef3b04562ace166a974a96ccbaa05c4bfdc85add2a621added6754e75717d9e"
            "0669929c316bd7a6cd4688526f4fa446ffa25457d9d50cf089e71d111eaed9b0"
        )
        instance.sign_message.return_value = mock_signature

        yield mock_account


class TestGenerateCommand:
    """Tests for the 'generate' command."""

    def test_generate_with_env_vars(
        self, runner: CliRunner, mock_account_setup: MagicMock
    ) -> None:
        """Test key generation using environment variables."""
        with patch.dict(
            "os.environ",
            {
                "GOOGLE_CLOUD_PROJECT": "test-project",
                "GOOGLE_CLOUD_REGION": "test-region",
                "KEY_RING": "test-keyring",
                "KEY_NAME": "test-key",
            },
            clear=True,  # Important: clear other env vars
        ):
            result = runner.invoke(app, ["generate"])

            assert result.exit_code == 0
            assert "Created Ethereum signing key" in result.stdout
            assert "0x0545640A0EcD6FB6ae94766811F30dCDA4746DFC" in result.stdout

    def test_generate_with_explicit_args(
        self, runner: CliRunner, mock_account_setup: MagicMock
    ) -> None:
        """Test key generation using explicit command line arguments."""
        result = runner.invoke(
            app,
            [
                "generate",
                "--project-id", "cli-project",
                "--location", "cli-region",
                "--keyring", "cli-keyring",
                "--key-id", "cli-key",
                "--retention-days", "180",
            ],
        )

        assert result.exit_code == 0
        assert "Created Ethereum signing key" in result.stdout

    def test_fail_generate_missing_required_args(self, runner: CliRunner, monkeypatch) -> None:
        """Test key generation fails when required arguments are missing."""
        # ? Need to clrear the env vars in order to raise error in the cli
        env_vars_to_clear = [
            "GOOGLE_CLOUD_PROJECT",
            "GOOGLE_CLOUD_REGION",
            "KEY_RING",
            "KEY_NAME",
            "GOOGLE_APPLICATION_CREDENTIALS",
            "GCP_ADC_CREDENTIALS_STRING"
        ]

        for env_var in env_vars_to_clear:
            monkeypatch.delenv(env_var, raising=False)
        result = runner.invoke(app, ["generate"])
        assert result.exit_code == 2
        assert "Usage:" in result.stdout

    #? These tests are skipped because of `memoryview: a bytes-like object is required, not 'MagicMock'`
    # def test_generate_handles_error(
    #     self, runner: CliRunner, mock_account_setup: MagicMock
    # ) -> None:
    #     """
    #     Test error handling during key generation by simulating a failure
    #     in the key creation process.
    #     """
    #     # Setup the error at the correct level in the dependency chain
    #     with patch("web3_google_hsm.accounts.gcp_kms_account.GCPKmsAccount.create_eth_key") as mock_create:
    #         # Make the create_eth_key method raise our test error
    #         mock_create.side_effect = Exception("Test error")

    #         # Execute the command with environment variables
    #         with patch.dict(
    #             "os.environ",
    #             {
    #                 "GOOGLE_CLOUD_PROJECT": "test-project",
    #                 "GOOGLE_CLOUD_REGION": "test-region",
    #                 "KEY_RING": "test-keyring",
    #                 "KEY_NAME": "test-key",
    #             },
    #             clear=True,
    #         ):
    #             result = runner.invoke(app, ["generate"])

    #             # For debugging, let's see what actually happened
    #             print(f"Command output: {result.stdout}")

    #             # Verify the error handling
    #             assert result.exit_code == 1
    #             assert "Test error" in result.stdout


class TestSignCommand:
    """Tests for the 'sign' command."""

    # def test_sign_message(
    #     self, runner: CliRunner, mock_account_setup: MagicMock
    # ) -> None:
    #     """Test signing a message with correct account."""
    #     with patch.dict(
    #         "os.environ",
    #         {
    #             "GOOGLE_CLOUD_PROJECT": "test-project",
    #             "GOOGLE_CLOUD_REGION": "test-region",
    #             "KEY_RING": "test-keyring",
    #             "KEY_NAME": "test-key",
    #         },
    #         clear=True,
    #     ):
    #         result = runner.invoke(
    #             app,
    #             [
    #                 "sign",
    #                 "Hello Ethereum!",
    #                 "--account", "0x0545640A0EcD6FB6ae94766811F30dCDA4746DFC",
    #             ],
    #         )
    #         print(result)
    #         assert result.exit_code == 0
    #         assert "Message signed successfully!" in result.stdout
    #         assert "Hello Ethereum!" in result.stdout

    def test_fail_sign_with_wrong_account(
        self, runner: CliRunner, mock_account_setup: MagicMock
    ) -> None:
        """Test signing fails with wrong account address."""
        with patch.dict(
            "os.environ",
            {
                "GOOGLE_CLOUD_PROJECT": "test-project",
                "GOOGLE_CLOUD_REGION": "test-region",
                "KEY_RING": "test-keyring",
                "KEY_NAME": "test-key",
            },
            clear=True,
        ):
            result = runner.invoke(
                app,
                [
                    "sign",
                    "Hello Ethereum!",
                    "--account", "0x0000000000000000000000000000000000000000",
                ],
            )

            assert result.exit_code == 1
            assert "Account mismatch" in result.stdout

    # def test_fail_sign_handles_signing_error(
    #     self, runner: CliRunner, mock_account_setup: MagicMock
    # ) -> None:
    #     """Test error handling during signing."""
    #     mock_account_setup.return_value.sign_message.side_effect = Exception(
    #         "Signing failed"
    #     )

    #     with patch.dict(
    #         "os.environ",
    #         {
    #             "GOOGLE_CLOUD_PROJECT": "test-project",
    #             "GOOGLE_CLOUD_REGION": "test-region",
    #             "KEY_RING": "test-keyring",
    #             "KEY_NAME": "test-key",
    #         },
    #         clear=True,
    #     ):
    #         result = runner.invoke(
    #             app,
    #             [
    #                 "sign",
    #                 "Hello Ethereum!",
    #                 "--account", "0x0545640A0EcD6FB6ae94766811F30dCDA4746DFC",
    #             ],
    #         )

    #         assert result.exit_code == 1
    #         assert "Signing failed" in result.stdout



"""Ethereum type definitions."""

from eth_account import Account
from eth_account._utils.legacy_transactions import (
    encode_transaction,  # noqa: PLC2701
    serializable_unsigned_transaction_from_dict,  # noqa: PLC2701
)
from eth_utils import is_address, to_checksum_address, to_int
from pydantic import BaseModel, ConfigDict, Field, field_validator

from web3_google_hsm.exceptions import SignatureError

SIGNATURE_LENGTH: int = 65
MSG_HASH_LENGTH: int = 32


class Signature(BaseModel):
    """Represents an Ethereum signature with v, r, s components."""

    v: int = Field(..., description="Recovery identifier")
    r: bytes = Field(..., description="R component of signature")
    s: bytes = Field(..., description="S component of signature")

    @field_validator("r", "s")
    @classmethod
    def validate_length(cls, v: bytes) -> bytes:
        if len(v) != MSG_HASH_LENGTH:
            msg = f"Length must be 32 bytes, got {len(v)} bytes"
            raise ValueError(msg)
        return v

    @field_validator("v")
    @classmethod
    def validate_v(cls, v: int) -> int:
        if v < 0:
            msg = "v must be non-negative"
            raise ValueError(msg)
        return v

    def to_hex(self) -> str:
        """Convert signature to hex string."""
        return "0x" + (self.r + self.s + bytes([self.v])).hex()

    @classmethod
    def from_hex(cls, hex_str: str) -> "Signature":
        """Create signature from hex string."""
        if hex_str.startswith("0x"):
            hex_str = hex_str[2:]
        sig_bytes = bytes.fromhex(hex_str)
        if len(sig_bytes) != SIGNATURE_LENGTH:
            msg = f"Invalid signature length: {len(sig_bytes)}"
            raise ValueError(msg)
        return cls(v=sig_bytes[64], r=sig_bytes[0:32], s=sig_bytes[32:64])


class Transaction(BaseModel):
    """Represents an Ethereum transaction."""

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    chain_id: int = Field(..., description="Chain ID", validation_alias="chainId")
    nonce: int = Field(..., ge=0, description="Transaction nonce")
    gas_price: int = Field(..., gt=0, description="Gas price in Wei", validation_alias="gasPrice")
    gas_limit: int = Field(..., gt=0, description="Gas limit", validation_alias="gas")
    to: str = Field(..., description="Recipient address")
    value: int = Field(..., ge=0, description="Transaction value in Wei")
    data: str = Field("0x", description="Transaction data")
    from_: str = Field(..., description="Sender address", validation_alias="from")
    signature: Signature | None = Field(None, description="Transaction signature")

    @field_validator("to", "from_")
    @classmethod
    def validate_address(cls, v: str) -> str:
        if not is_address(v):
            msg = "Invalid Ethereum address"
            raise ValueError(msg)
        return to_checksum_address(v)

    @field_validator("data")
    @classmethod
    def validate_hex(cls, v: str) -> str:
        if not v.startswith("0x"):
            v = "0x" + v
        try:
            bytes.fromhex(v[2:])
        except ValueError as error:
            msg = "Invalid hex string"
            raise ValueError(msg) from error
        return v

    def to_dict(self) -> dict:
        """Convert transaction to dictionary format for web3.py."""
        tx_dict = {
            "chainId": self.chain_id,
            "nonce": self.nonce,
            "gasPrice": self.gas_price,
            "gas": self.gas_limit,
            "to": self.to,
            "value": self.value,
            "data": self.data,
            "from": self.from_,
        }
        return tx_dict

    def to_transaction_dict(self) -> dict:
        """Convert to dictionary format suitable for signing."""
        tx_dict = {
            "chainId": self.chain_id,
            "nonce": self.nonce,
            "gasPrice": self.gas_price,
            "gas": self.gas_limit,
            "to": self.to,
            "value": self.value,
            "data": self.data,
        }
        # Add signature if present
        if self.signature:
            tx_dict.update(
                {
                    "v": self.signature.v,
                    "r": int.from_bytes(self.signature.r, "big"),
                    "s": int.from_bytes(self.signature.s, "big"),
                }
            )
        return tx_dict

    @classmethod
    def from_dict(cls, data: dict) -> "Transaction":
        """
        Create transaction from dictionary.

        Args:
            data: Transaction data dictionary.

        Returns:
            Transaction: A new transaction instance

        Raises:
            ValueError: If required fields are missing or invalid
        """
        tx_data = data.copy()

        # Handle signature if present
        signature = None
        if all(k in tx_data for k in ["v", "r", "s"]):
            r_value = tx_data.pop("r")
            s_value = tx_data.pop("s")

            # Convert hex strings to bytes if necessary
            if isinstance(r_value, str):
                r_value = bytes.fromhex(r_value[2:] if r_value.startswith("0x") else r_value)
            if isinstance(s_value, str):
                s_value = bytes.fromhex(s_value[2:] if s_value.startswith("0x") else s_value)

            signature = Signature(v=tx_data.pop("v"), r=r_value, s=s_value)
            tx_data["signature"] = signature

        return cls(**tx_data)

    def serialize_transaction(self) -> bytes:
        """
        Serialize a transaction to bytes.

        Returns:
            bytes: The serialized transaction

        Raises:
            SignatureError: If transaction is not signed or signature verification fails
        """
        if not self.signature:
            msg = "The transaction is not signed."
            raise SignatureError(msg)

        # Create transaction dict without 'from' and with proper signature format
        txn_data = self.to_dict()

        if "from" in txn_data:
            txn_data.pop("from")

        # Create unsigned transaction dict
        unsigned_txn = serializable_unsigned_transaction_from_dict(txn_data)
        signature = (self.signature.v, to_int(self.signature.r), to_int(self.signature.s))

        signed_txn = encode_transaction(unsigned_txn, signature)

        # Verify signature
        recovered = Account.recover_transaction(signed_txn)
        if self.from_ and recovered.lower() != self.from_.lower():
            msg = f"Recovered signer doesn't match sender! Expected: {self.from_}, got: {recovered}"
            raise SignatureError(msg)

        return signed_txn



import os

import pytest
from unittest.mock import MagicMock

from web3_google_hsm.accounts.gcp_kms_account import GCPKmsAccount
from web3_google_hsm import Transaction
from eth_account.messages import encode_defunct
from hexbytes import HexBytes


# Check required environment variables
REQUIRED_ENV_VARS = {
    "GOOGLE_CLOUD_PROJECT": os.getenv("GOOGLE_CLOUD_PROJECT"),
    "GOOGLE_CLOUD_REGION": os.getenv("GOOGLE_CLOUD_REGION"),
    "KEY_RING": os.getenv("KEY_RING"),
    "KEY_NAME": os.getenv("KEY_NAME"),
    "JSON_RPC_BASE": os.getenv("JSON_RPC_BASE", "http://localhost:8545"),
}

# Skip all tests if any required env var is missing
missing_vars = [k for k, v in REQUIRED_ENV_VARS.items() if not v]
pytestmark = pytest.mark.skipif(
    bool(missing_vars),
    reason=f"Missing required environment variables: {', '.join(missing_vars)}"
)

def test_account_initialization(gcp_account):
    """Test initializing account with real GCP KMS."""
    assert gcp_account.key_path
    assert gcp_account.address.startswith("0x")
    assert len(gcp_account.address) == 42


def test_message_signing_and_verification(gcp_account, web3):
    """Test signing and verifying messages with real GCP KMS."""
    # Sign message
    message = "Hello Ethereum!"
    signature = gcp_account.sign_message(message)

    # Verify components
    assert signature.v in (27, 28)
    assert len(signature.r) == 32
    assert len(signature.s) == 32

    # Verify recovery
    message_hash = encode_defunct(text=message)
    recovered_address = web3.eth.account.recover_message(
        message_hash,
        vrs=(signature.v, signature.r, signature.s)
    )

    assert recovered_address.lower() == gcp_account.address.lower()


def test_transaction_signing(gcp_account, fund_account, web3):
    """Test signing and sending transactions with real GCP KMS."""
    # Create transaction
    tx = Transaction(
        chain_id=web3.eth.chain_id,
        nonce=web3.eth.get_transaction_count(gcp_account.address),
        gas_price=web3.eth.gas_price,
        gas_limit=21000,
        to="0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
        value=web3.to_wei(0.0001, "ether"),
        data="0x",
        from_=gcp_account.address
    )

    # Sign transaction
    signed_tx = gcp_account.sign_transaction(tx)
    assert signed_tx is not None

    # Send transaction
    tx_hash = web3.eth.send_raw_transaction(signed_tx)
    receipt = web3.eth.wait_for_transaction_receipt(tx_hash)

    # Verify transaction was successful
    assert receipt['status'] == 1
    assert receipt['from'].lower() == gcp_account.address.lower()

    # Verify signature
    recovered = web3.eth.account.recover_transaction(signed_tx)
    assert recovered.lower() == gcp_account.address.lower()


def test_transaction_with_data(gcp_account, fund_account, web3):
    """Test signing transactions with data field."""
    tx = Transaction(
        chain_id=web3.eth.chain_id,
        nonce=web3.eth.get_transaction_count(gcp_account.address),
        gas_price=300000000000,
        gas_limit=1000000,
        to="0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
        value=web3.to_wei(0.000001, "ether"),
        data="0x68656c6c6f",  # "hello" in hex
        from_=gcp_account.address
    )

    signed_tx = gcp_account.sign_transaction(tx)
    assert signed_tx is not None

    # Send and verify
    tx_hash = web3.eth.send_raw_transaction(signed_tx)
    receipt = web3.eth.wait_for_transaction_receipt(tx_hash)
    assert receipt['status'] == 1

    # Verify the transaction data
    tx_details = web3.eth.get_transaction(tx_hash)
    assert tx_details['input'] == HexBytes("0x68656c6c6f")


def test_multiple_transactions(gcp_account, web3):
    """Test sending multiple consecutive transactions."""
    for i in range(3):
        tx = Transaction(
            chain_id=web3.eth.chain_id,
            nonce=web3.eth.get_transaction_count(gcp_account.address),
            gas_price=web3.eth.gas_price,
            gas_limit=21000,
            to="0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
            value=web3.to_wei(0.0001, "ether"),
            data="0x",
            from_=gcp_account.address
        )

        signed_tx = gcp_account.sign_transaction(tx)
        tx_hash = web3.eth.send_raw_transaction(signed_tx)
        receipt = web3.eth.wait_for_transaction_receipt(tx_hash)

        assert receipt['status'] == 1
        assert receipt['from'].lower() == gcp_account.address.lower()


def test_account_initialization(gcp_kms_account: GCPKmsAccount):
    """Test account initialization."""
    assert isinstance(gcp_kms_account, GCPKmsAccount)
    assert gcp_kms_account.key_path is not None

def test_get_public_key(gcp_kms_account: GCPKmsAccount, mock_kms_client: MagicMock):
    """Test getting public key from KMS."""
    public_key = gcp_kms_account.public_key
    assert isinstance(public_key, bytes)
    assert len(public_key) == 64
    mock_kms_client.get_public_key.assert_called_once()

def test_get_address(gcp_kms_account: GCPKmsAccount, test_address: str):
    """Test deriving Ethereum address from public key."""
    address = gcp_kms_account.address
    assert address == test_address
    assert address.startswith("0x")
    assert len(address) == 42


def test_invalid_message_type(gcp_kms_account: GCPKmsAccount):
    """Test signing with invalid message type."""
    with pytest.raises(TypeError, match="Unsupported message type"):
        gcp_kms_account.sign_message(123)  # type: ignore

def test_sign_transaction_no_signature(gcp_kms_account: GCPKmsAccount, mock_kms_client: MagicMock):
    """Test transaction signing when KMS fails."""
    mock_kms_client.asymmetric_sign.return_value = None

    with pytest.raises(Exception, match="Signing error"):
        gcp_kms_account.sign_message("test message")



"""CLI interface for web3-google-hsm."""

import typer
from rich.console import Console

from web3_google_hsm.accounts.gcp_kms_account import GCPKmsAccount
from web3_google_hsm.config import BaseConfig

console = Console()
app = typer.Typer(help="🔐 Web3 Google HSM CLI tools for Ethereum key management and signing.")


@app.command()
def generate(
    project_id: str = typer.Option(
        ...,
        envvar="GOOGLE_CLOUD_PROJECT",
        help="Google Cloud project ID",
    ),
    location: str = typer.Option(
        ...,
        envvar="GOOGLE_CLOUD_REGION",
        help="Cloud KMS location (e.g. us-east1)",
    ),
    keyring: str = typer.Option(
        ...,
        envvar="KEY_RING",
        help="Name of the key ring",
    ),
    key_id: str = typer.Option(
        ...,
        envvar="KEY_NAME",
        help="ID for the new key",
    ),
    retention_days: int = typer.Option(
        365,
        help="Days to retain key versions",
    ),
) -> None:
    """🔑 Generate a new Ethereum signing key in Cloud HSM."""
    try:
        # Create the key
        key = GCPKmsAccount.create_eth_key(
            project_id=project_id,
            location_id=location,
            key_ring_id=keyring,
            key_id=key_id,
            retention_days=retention_days,
        )
        console.print(f"[green]✅ Created Ethereum signing key:[/green] {key.name}")

        # Initialize account to display the Ethereum address
        config = BaseConfig(
            project_id=project_id,
            location_id=location,
            key_ring_id=keyring,
            key_id=key_id,
        )
        account = GCPKmsAccount(config=config)
        console.print(f"[blue]🔑 Ethereum address:[/blue] {account.address}")

    except Exception as e:
        console.print(f"[red]❌ Error:[/red] {e!s}")
        raise typer.Exit(1) from e


@app.command()
def sign(
    message: str = typer.Argument(
        ...,
        help="Message to sign",
    ),
    account: str = typer.Option(
        ...,
        "--account",
        "-a",
        help="Ethereum address of the signing account",
    ),
) -> None:
    """📝 Sign a message using a Cloud HSM key."""
    try:
        # Initialize from environment variables
        config = BaseConfig.from_env()
        signer = GCPKmsAccount(config=config)

        # Verify the account matches
        if signer.address.lower() != account.lower():
            console.print(f"[red]❌ Account mismatch:[/red] {signer.address} != {account}")
            raise typer.Exit(1)

        # Sign the message
        signature = signer.sign_message(message)

        console.print("[green]✅ Message signed successfully![/green]")
        console.print(f"[blue]📝 Message:[/blue] {message}")
        console.print(f"[blue]🔏 Signature:[/blue] {signature.to_hex()}")
        console.print("[blue]📊 Components:[/blue]")
        console.print(f"  v: {signature.v}")
        console.print(f"  r: 0x{signature.r.hex()}")
        console.print(f"  s: 0x{signature.s.hex()}")

    except Exception as e:
        console.print(f"[red]❌ Error:[/red] {e!s}")
        raise typer.Exit(1) from e


def cli() -> None:
    """Entry point for the CLI."""
    app()


if __name__ == "__main__":
    cli()



# from typing import Any

import ecdsa  # type: ignore

# from ape.api.transactions import TransactionAPI
# from ape.types import MessageSignature
from cryptography.hazmat.primitives import serialization

# from eth_account.messages import encode_defunct
# from web3 import Web3

SECP256_K1_N = int("fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141", 16)


def extract_public_key_bytes(pem_str: str) -> bytes:
    """
    Extract raw public key bytes from PEM format.

    Args:
        pem_str: PEM-encoded public key string

    Returns:
        bytes: Raw public key bytes (64 bytes of X,Y coordinates)

    Raises:
        ValueError: If the PEM string is invalid
    """
    if not pem_str.startswith("-----BEGIN PUBLIC KEY-----"):
        pem_str = f"-----BEGIN PUBLIC KEY-----\n{pem_str}\n-----END PUBLIC KEY-----"

    try:
        pem_bytes = pem_str.encode()
        public_key = serialization.load_pem_public_key(pem_bytes)

        raw_bytes = public_key.public_bytes(
            encoding=serialization.Encoding.X962, format=serialization.PublicFormat.UncompressedPoint
        )
        return raw_bytes[-64:]
    except Exception as err:
        msg = "Invalid PEM format"
        raise ValueError(msg) from err


def convert_der_to_rsv(signature: bytes, v_adjustment_factor: int = 0) -> dict:
    """
    Convert DER signature to RSV format.

    Args:
        signature: The DER-encoded signature bytes
        v_adjustment_factor: The v value adjustment factor for the signature

    Returns:
        SignatureComponents: A dictionary containing the r, s, and v components

    Raises:
        ValueError: If the signature is invalid or cannot be decoded
    """
    r, s = ecdsa.util.sigdecode_der(signature, ecdsa.SECP256k1.order)
    v = v_adjustment_factor
    if s > SECP256_K1_N / 2:
        s = SECP256_K1_N - s
    r = r.to_bytes(32, byteorder="big")
    s = s.to_bytes(32, byteorder="big")
    return {"v": v, "r": r, "s": s}


# def verify_signed_message(
#     w3: Web3, message: str, signature: MessageSignature, signer_address: str
# ) -> tuple[bool, dict[str, Any]]:
#     """
#     Verify a signed Ethereum message.
#
#     Args:
#         w3: Web3 instance
#         message: Original message that was signed
#         signature: MessageSignature object containing v, r, s
#         signer_address: Expected signer's address
#
#     Returns:
#         Tuple[bool, dict]: (is_valid, details)
#
#     Example:
#         >>> w3 = Web3(Web3.HTTPProvider('http://localhost:8545'))
#         >>> signed_message = account.sign_message("Hello")
#         >>> is_valid, details = verify_signed_message(
#         ...     w3, "Hello", signed_message, account.address
#         ... )
#     """
#     try:
#         message_hash = encode_defunct(text=message)
#         recovered_address = w3.eth.account.recover_message(message_hash, vrs=(signature.v, signature.r, signature.s))
#
#         is_valid = recovered_address.lower() == signer_address.lower()
#
#         return is_valid, {
#             "original_address": signer_address,
#             "recovered_address": recovered_address,
#             "is_valid": is_valid,
#             "signature": {
#                 "r": signature.r.hex(),
#                 "s": signature.s.hex(),
#                 "v": signature.v,
#                 "full_signature": signature.encode_vrs().hex(),
#             },
#         }
#     except Exception as e:
#         return False, {"error": str(e)}
#
#
# def verify_signed_transaction(w3: Web3, signed_tx: TransactionAPI, signer_address: str) -> tuple[bool, dict[str, Any]]:  # noqa: E501
#     """
#     Verify a signed Ethereum transaction.
#
#     Args:
#         w3: Web3 instance
#         signed_tx: Signed transaction
#         signer_address: Expected signer's address
#
#     Returns:
#         Tuple[bool, dict]: (is_valid, details)
#
#     Example:
#         >>> w3 = Web3(Web3.HTTPProvider('http://localhost:8545'))
#         >>> signed_tx = account.sign_transaction(StaticFeeTransaction(**tx))
#         >>> is_valid, details = verify_signed_transaction(
#         ...     w3, signed_tx, account.address
#         ... )
#     """
#     try:
#         raw_tx = signed_tx.serialize_transaction()
#         recovered_address = w3.eth.account.recover_transaction(raw_tx)
#
#         is_valid = recovered_address.lower() == signer_address.lower()
#
#         return is_valid, {
#             "original_address": signer_address,
#             "recovered_address": recovered_address,
#             "is_valid": is_valid,
#             "transaction_hash": w3.keccak(raw_tx).hex(),
#         }
#     except Exception as e:
#         return False, {"error": str(e)}



import dotenv
from eth_account.messages import encode_defunct
from web3 import Web3

from web3_google_hsm.accounts.gcp_kms_account import GCPKmsAccount
from web3_google_hsm.types.ethereum_types import Transaction

if __name__ == "__main__":
    from rich.console import Console

    console = Console()
    dotenv.load_dotenv()
    w3 = Web3(Web3.HTTPProvider("http://localhost:8545"))
    account = GCPKmsAccount()

    console.print(f"GCP KMS Account address: {account.address}")

    # 1. Test Message Signing and Verification
    message = "Hello Ethereum!"
    message_hash = encode_defunct(text=message)

    # Sign the message
    signed_message = account.sign_message(message)
    console.print("\nSigned message details:")
    console.print(f"R: {signed_message.r.hex()}")
    console.print(f"S: {signed_message.s.hex()}")
    console.print(f"V: {signed_message.v}")
    console.print(f"Full signature: {signed_message.to_hex()}")

    # Verify the signature using web3.py
    recovered_address = w3.eth.account.recover_message(
        message_hash, vrs=(signed_message.v, signed_message.r, signed_message.s)
    )
    console.print("\nSignature verification:")
    console.print(f"Original address: {account.address}")
    console.print(f"Recovered address: {recovered_address}")
    console.print(f"Signature valid: {recovered_address.lower() == account.address.lower()}")

    # 2. Test Transaction Signing and Sending
    console.print("\nTesting transaction signing...")

    # First, fund the account from a test account (using Anvil's default funded account)
    funded_account = w3.eth.account.from_key("0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80")

    # Send some ETH to our GCP KMS account
    fund_tx = {
        "from": funded_account.address,
        "to": account.address,
        "value": w3.to_wei(10, "ether"),
        "gas": 21000,
        "gasPrice": w3.eth.gas_price,
        "nonce": w3.eth.get_transaction_count(funded_account.address),
        "chainId": w3.eth.chain_id,
    }

    # Send funding transaction
    signed_fund_tx = w3.eth.account.sign_transaction(fund_tx, funded_account.key)
    fund_tx_hash = w3.eth.send_raw_transaction(signed_fund_tx.raw_transaction)
    fund_receipt = w3.eth.wait_for_transaction_receipt(fund_tx_hash)
    console.print(f"Funded account with 10 ETH. TX hash: {fund_receipt['transactionHash'].hex()}")

    # Now create and sign a transaction from our GCP KMS account
    tx = {
        "from": funded_account.address,
        "chain_id": w3.eth.chain_id,
        "nonce": w3.eth.get_transaction_count(account.address),
        "value": w3.to_wei(0.000001, "ether"),
        "data": "0x00",
        "to": "0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
        "type": 0,
        "gas_limit": 1000000,
        "gas_price": 300000000000,
    }

    # Sign the transaction
    signed_tx = account.sign_transaction(Transaction.from_dict(tx))
    console.print(f"{signed_tx=}")

    if signed_tx:
        console.print(signed_tx)
        # Send the transaction
        tx_hash = w3.eth.send_raw_transaction(signed_tx)
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        console.print("\nTransaction successful!")
        console.print(f"Transaction hash: {receipt['transactionHash'].hex()}")
        console.print(f"From: {receipt['from']}")
        console.print(f"To: {receipt['to']}")
        console.print(f"Gas used: {receipt['gasUsed']}")

        # Verify the transaction signature
        tx_data = w3.eth.get_transaction(tx_hash)
        # Get the raw transaction data
        raw_tx = signed_tx
        recovered_address = w3.eth.account.recover_transaction(raw_tx)
        console.print("\nTransaction signature verification:")
        console.print(f"Original address: {account.address}")
        console.print(f"Recovered address: {recovered_address}")
        console.print(f"Signature valid: {recovered_address.lower() == account.address.lower()}")
    else:
        console.print("Failed to sign transaction")



import os

from pydantic import ValidationError
import pytest
from web3_google_hsm.accounts.gcp_kms_account import GCPKmsAccount
from web3_google_hsm.config import BaseConfig
import json

# Define required environment variables
REQUIRED_ENV_VARS = {
    "GOOGLE_CLOUD_PROJECT": os.getenv("GOOGLE_CLOUD_PROJECT"),
    "GOOGLE_CLOUD_REGION": os.getenv("GOOGLE_CLOUD_REGION"),
    "KEY_RING": os.getenv("KEY_RING"),
    "KEY_NAME": os.getenv("KEY_NAME"),
    "GCP_ADC_CREDENTIALS_STRING": os.getenv("GCP_ADC_CREDENTIALS_STRING"),
}

# Skip all tests if any required env var is missing
missing_vars = [k for k, v in REQUIRED_ENV_VARS.items() if not v]
pytestmark = pytest.mark.skipif(
    bool(missing_vars),
    reason=f"Missing required environment variables: {', '.join(missing_vars)}"
)

def test_account_initialization_with_both():
    """Test initializing account with both config and credentials."""
    # Load credentials from GCP_ADC_CREDENTIALS_STRING env var
    credentials = json.loads(os.environ["GCP_ADC_CREDENTIALS_STRING"])

    # Create config from environment
    config = BaseConfig.from_env()

    # Initialize account with both
    account = GCPKmsAccount(config=config, credentials=credentials)

    # Verify initialization
    assert account._client is not None
    assert account._settings is not None
    assert account.key_path is not None
    assert account.address.startswith("0x")

    # Test basic functionality
    message = "Test message"
    signature = account.sign_message(message)
    assert signature.v in (27, 28)
    assert len(signature.r) == 32
    assert len(signature.s) == 32

def test_account_initialization_with_neither():
    """Test initializing account with neither config nor credentials (using env vars)."""
    # Initialize account without explicit config or credentials
    account = GCPKmsAccount()

    # Verify initialization
    assert account._client is not None
    assert account._settings is not None  # Should be created from env
    assert account.key_path is not None
    assert account.address.startswith("0x")

    # Test basic functionality
    message = "Test message"
    signature = account.sign_message(message)
    assert signature.v in (27, 28)
    assert len(signature.r) == 32
    assert len(signature.s) == 32

def test_fail_account_initialization_with_only_config(monkeypatch):
    """Test that initializing with only config raises error."""
    env_vars_to_clear = [
        "GOOGLE_CLOUD_PROJECT",
        "GOOGLE_CLOUD_REGION",
        "KEY_RING",
        "KEY_NAME",
        "GOOGLE_APPLICATION_CREDENTIALS",
        "GCP_ADC_CREDENTIALS_STRING"
    ]

    for env_var in env_vars_to_clear:
        monkeypatch.delenv(env_var, raising=False)

    with pytest.raises(ValidationError):
        config = BaseConfig.from_env()
        GCPKmsAccount(config=config)

def test_fail_account_initialization_with_only_credentials(monkeypatch):
    """Test that initializing with only credentials raises error."""
    env_vars_to_clear = [
        "GOOGLE_CLOUD_PROJECT",
        "GOOGLE_CLOUD_REGION",
        "KEY_RING",
        "KEY_NAME",
    ]

    for env_var in env_vars_to_clear:
        monkeypatch.delenv(env_var, raising=False)
    credentials = json.loads(os.environ["GCP_ADC_CREDENTIALS_STRING"])

    with pytest.raises(ValidationError):
        GCPKmsAccount(credentials=credentials)

def test_key_path_matches_config(monkeypatch):
    """Test that the key path matches the config values."""
    # Load both config and credentials
    credentials = json.loads(os.environ["GCP_ADC_CREDENTIALS_STRING"])

    config = BaseConfig.from_env()
    account = GCPKmsAccount(config=config, credentials=credentials)

    # Verify key path contains all the expected components
    assert config.project_id in account.key_path
    assert config.location_id in account.key_path
    assert config.key_ring_id in account.key_path
    assert config.key_id in account.key_path



import datetime
from functools import cached_property
from typing import Any, cast

import rlp  # type: ignore
from eth_account import Account
from eth_account._utils.legacy_transactions import serializable_unsigned_transaction_from_dict  # noqa: PLC2701
from eth_account.messages import _hash_eip191_message, encode_defunct  # noqa: PLC2701
from eth_typing import ChecksumAddress
from eth_utils import keccak, to_checksum_address
from google.auth import load_credentials_from_dict
from google.cloud import kms
from google.protobuf import duration_pb2  # type: ignore
from pydantic import BaseModel, Field, PrivateAttr
from rich.traceback import install

from web3_google_hsm.config import BaseConfig
from web3_google_hsm.exceptions import SignatureError
from web3_google_hsm.types.ethereum_types import MSG_HASH_LENGTH, Signature, Transaction
from web3_google_hsm.utils import convert_der_to_rsv, extract_public_key_bytes

# initialise the rich traceback for better tracebacks
install()


class GCPKmsAccount(BaseModel):
    """Account implementation using Google Cloud KMS."""

    # Public fields
    key_path: str = Field(default="")

    # Private attributes
    _client: kms.KeyManagementServiceClient = PrivateAttr()
    _cached_public_key: bytes | None = PrivateAttr(default=None)
    _settings: BaseConfig = PrivateAttr()

    def __init__(self, config: BaseConfig | None = None, credentials: dict | None = None, **data: Any):
        """
        Initialize GCP KMS Account with either config or credentials.
        If neither is provided, uses Google SDK default auth mechanism.

        Args:
            config: BaseConfig instance for environment-based configuration
            credentials: Dictionary containing GCP credentials
            **data: Additional data passed to BaseModel

        Raises:
            ValueError: If both config and credentials are provided
        """

        super().__init__(**data)

        if isinstance(credentials, dict):
            credentials, _ = load_credentials_from_dict(credentials)
        # Initialize client based on provided auth method
        self._client = (
            kms.KeyManagementServiceClient(credentials=credentials) if credentials else kms.KeyManagementServiceClient()  # type: ignore
        )

        # Initialize settings if config is provided, otherwise None
        self._settings = config or BaseConfig.from_env()

        # Set key path based on config or credentials
        self.key_path = self._get_key_version_path()

    def _get_key_version_path(self) -> str:
        """Get the full path to the key version in Cloud KMS."""
        return self._client.crypto_key_version_path(
            self._settings.project_id,
            self._settings.location_id,
            self._settings.key_ring_id,
            self._settings.key_id,
            "1",  # Using version 1
        )

    @property
    def public_key(self) -> bytes:
        """Get public key bytes from KMS."""
        if self._cached_public_key is None:
            response = self._client.get_public_key({"name": self.key_path})
            if not response.pem:
                msg = "No PEM data in response"
                raise ValueError(msg)

            self._cached_public_key = extract_public_key_bytes(response.pem)
        return self._cached_public_key

    @cached_property
    def address(self) -> ChecksumAddress:
        """Get Ethereum address derived from public key."""
        return to_checksum_address(keccak(self.public_key)[-20:].hex().lower())

    @classmethod
    def create_eth_key(
        cls,
        project_id: str,
        location_id: str,
        key_ring_id: str,
        key_id: str,
        retention_days: int = 365,
    ) -> kms.CryptoKey:
        """
        Creates a new Ethereum signing key in Cloud KMS backed by Cloud HSM.

        Args:
            project_id: Google Cloud project ID
            location_id: Cloud KMS location (e.g. 'us-east1')
            key_ring_id: ID of the Cloud KMS key ring
            key_id: ID of the key to create
            retention_days: Days to retain key versions before destruction (default: 365)

        Returns:
            CryptoKey: Created Cloud KMS key

        Raises:
            Exception: If key creation fails
        Reference:
            https://github.com/GoogleCloudPlatform/python-docs-samples/blob/main/kms/snippets/create_key_hsm.py
        """
        try:
            client = kms.KeyManagementServiceClient()
            key_ring_name = client.key_ring_path(project_id, location_id, key_ring_id)

            # Configure for Ethereum signing
            purpose = kms.CryptoKey.CryptoKeyPurpose.ASYMMETRIC_SIGN
            algorithm = kms.CryptoKeyVersion.CryptoKeyVersionAlgorithm.EC_SIGN_SECP256K1_SHA256
            protection_level = kms.ProtectionLevel.HSM

            key = {
                "purpose": purpose,
                "version_template": {
                    "algorithm": algorithm,
                    "protection_level": protection_level,
                },
                "destroy_scheduled_duration": duration_pb2.Duration().FromTimedelta(
                    datetime.timedelta(days=retention_days)
                ),
            }

            return client.create_crypto_key(
                request={"parent": key_ring_name, "crypto_key_id": key_id, "crypto_key": key}
            )
        except Exception as e:
            msg = f"Failed to create key: {e}"
            raise Exception(msg) from e

    def _sign_raw_hash(self, msghash: bytes) -> bytes | None:
        """Sign a message hash using KMS."""
        try:
            response = self._client.asymmetric_sign(request={"name": self.key_path, "digest": {"sha256": msghash}})
            return response.signature
        except Exception as e:
            msg = f"Signing error: {e}"
            raise Exception(msg) from e

    def sign_message(self, message: str | bytes) -> Signature:
        """
        Sign a message with the GCP KMS key.

        Args:
            message: Message to sign (str or bytes)

        Returns:
            Signature: The v, r, s components of the signature

        Raises:
            TypeError: If message is not str or bytes
            ValueError: If message hash length is invalid
            SignatureError: If signature verification fails
            Exception: If signing fails

        Example:
            ```{ .python .copy }
                account = GCPKmsAccount()
                print(f"GCP KMS Account address: {account.address}")
                message = "Hello Ethereum!"
                # Sign the message
                signed_message = account.sign_message(message)
            ```
        """
        # Convert message to SignableMessage format
        if isinstance(message, str):
            if message.startswith("0x"):
                hash_message = encode_defunct(hexstr=message)
            else:
                hash_message = encode_defunct(text=message)
        elif isinstance(message, bytes):
            hash_message = encode_defunct(primitive=message)
        else:
            msg = f"Unsupported message type: {type(message)}"
            raise TypeError(msg)

        # Sign message hash
        msghash = _hash_eip191_message(hash_message)
        if len(msghash) != MSG_HASH_LENGTH:
            msg = "Invalid message hash length"
            raise ValueError(msg)

        der_signature = self._sign_raw_hash(msghash)
        if not der_signature:
            msg = "Failed to sign message"
            raise Exception(msg)

        # Try both v values (27 and 28) to find the correct one
        for v_value in (27, 28):
            sig_dict = convert_der_to_rsv(der_signature, v_value)
            signature = Signature(v=sig_dict["v"], r=sig_dict["r"], s=sig_dict["s"])

            # Verify the signature
            recovered = Account.recover_message(hash_message, vrs=(signature.v, signature.r, signature.s))

            if recovered.lower() == self.address.lower():
                return signature

        msg = "Failed to create valid signature"
        raise SignatureError(msg)

    def sign_transaction(self, transaction: Transaction) -> bytes | None:
        """
        Sign an EIP-155 transaction.

        Args:
            transaction: Transaction to sign

        Returns:
            bytes | None: Serialized signed transaction or None if signing fails
        """
        # Create unsigned transaction dictionary
        unsigned_tx = {
            "nonce": transaction.nonce,
            "gasPrice": transaction.gas_price,
            "gas": transaction.gas_limit,
            "to": transaction.to,
            "value": transaction.value,
            "data": transaction.data,
            "chainId": transaction.chain_id,
        }

        # Convert to UnsignedTransaction and get hash
        unsigned_tx_obj = serializable_unsigned_transaction_from_dict(unsigned_tx)  # type: ignore
        msg_hash = unsigned_tx_obj.hash()

        # Sign the transaction hash
        der_signature = self._sign_raw_hash(msg_hash)
        if not der_signature:
            return None

        # Calculate v value based on chain ID
        v_base = (2 * transaction.chain_id + 35) if transaction.chain_id else 27
        sig_dict = convert_der_to_rsv(der_signature, v_base)

        # Create RLP serializable fields
        rlp_data = [
            transaction.nonce,
            transaction.gas_price,
            transaction.gas_limit,
            bytes.fromhex(transaction.to[2:]),  # Convert address to bytes
            transaction.value,
            bytes.fromhex(transaction.data[2:] if transaction.data.startswith("0x") else transaction.data),
            sig_dict["v"],
            int.from_bytes(sig_dict["r"], "big"),
            int.from_bytes(sig_dict["s"], "big"),
        ]

        # RLP encode the transaction and ensure it returns bytes
        encoded_tx = cast(bytes, rlp.encode(rlp_data))

        # Verify the signature
        recovered = Account.recover_transaction(encoded_tx)
        if recovered.lower() != self.address.lower():
            # Try with v + 1
            rlp_data[6] = sig_dict["v"] + 1  # Update v value
            encoded_tx = cast(bytes, rlp.encode(rlp_data))

            # Verify again
            recovered = Account.recover_transaction(encoded_tx)
            if recovered.lower() != self.address.lower():
                msg = "Failed to create valid signature"
                raise SignatureError(msg)

        return encoded_tx
